/* ************************************************************************ 
 *         The Amulet User Interface Development Environment              *
 * ************************************************************************
 * This code was written as part of the Amulet project at                 *
 * Carnegie Mellon University, and has been placed in the public          *
 * domain.  If you are using this code or any part of Amulet,             *
 * please contact amulet@cs.cmu.edu to be put on the mailing list.        *
 * ************************************************************************/

/* David Mathog, Biology Division, Caltech
This variant can also output graphics to a file via Am_Save_To_GRAPHIC.
The graphic mode options (MUST be upper case) are: 

  POSTSCRIPT
  CHARACTER            these there are all CGM
  CLEAR_TEXT           "
  BINARY               "
  HPGL

This utilizes pieces from the RALCGM distribution, see the copyright
notices there.  Parts are nearly verbatim from TESTCGM.C.

WARNING! No commercial use without contacting RALCGM authors!!!!
*/

#ifdef GCC
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "cgmlib.h" // RALCGM include
#else
extern "C" {
#include <stdlib.h>
#include <limits.h> //INT_MIN
#include "cgmlib.h" // RALCGM include
#if defined(_MSC_VER) || defined(NEED_STRING)
#include <string.h>
#endif
}
#endif

#include <am_inc.h>

#include STANDARD_SLOTS__H
#include VALUE_LIST__H
#include GDEFS__H  //for Am_Point_List
#include WIDGETS__H //for Am_Get_Input_From_Dialog

#ifdef HPGL
   extern "C" void HPGLoptions ( char * );
#endif
#ifdef POSTSCRIPT
   extern "C" void PSoptions ( char * );
#endif

extern Am_Object window;
int cgm_save_y;

//
// conversions from amulet to 
//
#define to255(a) ((int)(255.0 * a))
#define cgm_rgb_color(col,a,b,c) {col.red=a;  col.green=b;   col.blue=c;}
#define cgm_col_color(col,i,a,b,c) {col.index=i; col.red=a; col.green=b; col.blue=c;}
#define cgm_set_point(a,b,c)  {a.x = ((float)b); a.y = ((float)(cgm_save_y - (int)c));} 
#define cgm_set_abspoint(a,b,c)  {a.x = ((float)b); a.y = ((float)c);}
#define nint(a) ( a<0 ? -((int)((-a) + 0.5)) : ((int)((a) + 0.5)))

// mode = 0 = graphics object other than line
//      = 1 = text object
//      = 2 = line
void set_gr_fill_edge_properties(Am_Object current, int mode){
float t_red,t_green,t_blue;
float fwidth;
int cd_fill_var;
int cd_hold_color;
Colour acolour;
short t1;
Am_Line_Cap_Style_Flag cap;

float fexpand[12] = {0.6, 0.7, 0.7, 0.7,   //empirically determined 
                     0.55, 0.55, 0.55, 0.55,   //expansion factors
                     0.8, 0.8, 0.8, 0.8};

// Always test that each type of slot exists.  Lines don't have fill,
// for instance. Some fancy program may have removed a standard slot
// from a graphics object

if(mode == 0 || mode == 2){
  if(current.Peek(Am_FILL_STYLE).Exists()){
    if(current.Get(Am_FILL_STYLE) == Am_No_Style){
      m_intstyle(HOLLOW);
      t_red = 0; t_green = 0; t_blue = 0;  // have to specify something
      }
    else {
      m_intstyle(SOLID);
      Am_Style (current.Get(Am_FILL_STYLE)).Get_Values (t_red, t_green, t_blue);  // fill color
      }
    cgm_col_color(acolour,1,to255(t_red),to255(t_green),to255(t_blue));
 
    m_fillcolr(acolour); //fill color as specified, or black for default
  }
//
//                    edge color, if any (no edge required)
//
  if(current.Peek(Am_LINE_STYLE).Exists()){
    if(current.Get(Am_LINE_STYLE) == Am_No_Style){
      m_edgevis(INVIS); //edge not visible
      t_red = 0; t_green = 0; t_blue = 0;  // have to specify something
      t1 = 1;
      }
    else {
      m_edgevis(VIS); //edge visible
      Am_Style (current.Get(Am_LINE_STYLE)).Get_Values (t_red, t_green, t_blue);  // fill color
      Am_Style (current.Get(Am_LINE_STYLE)).Get_Line_Thickness_Values (t1, cap);
      if(t1 <1 )t1 =1;
      }
    cgm_col_color(acolour,1,to255(t_red),to255(t_green),to255(t_blue));
    fwidth = t1;
    if (mode == 0){
       m_edgecolr(acolour); //edge color as specified, or black for default
       m_edgewidth(fwidth); //edge thickness
       }
    else {
       m_linecolr(acolour); //line color as specified, or black for default
       m_linewidth(fwidth); //line thickness
       }
    }
  }
else {
  if(current.Peek(Am_FONT).Exists()){
    bool is_bold, is_italic, is_underlined;
    Am_String fname;
    Am_Font_Family_Flag used_font;
    Am_Font_Size_Flag size;
    Am_Font (current.Get(Am_FONT)).Get(fname, used_font, is_bold, is_italic, is_underlined, size);
// set the font from the 12 we have defined (ignore family)
    int thisfont = 1;             //default = courier, goes with Am_FONT_FIXED
    if(used_font == Am_FONT_SERIF)thisfont = 2;
    if(used_font == Am_FONT_SANS_SERIF)thisfont = 3;
    int thisselect = 1 + 4 * (thisfont - 1);
// pick up the FIRST of bold, italic, or combination.  Underlined is not 
// supported, so set it to boldoblique.
    int thisoffset = 0;
    if(is_bold)thisoffset = 1;
    if(is_italic)thisoffset = thisoffset + 2;  //2 for italic, 3 for bold+italic
    if(is_underlined)thisoffset = 3; //no underline, use bold+italic
    m_textfontindex(thisselect + thisoffset); // one of three fonts
    m_charexpan(fexpand[(thisselect - 1)]);
  }
  if(current.Peek(Am_LINE_STYLE).Exists()){
    Am_Style (current.Get(Am_LINE_STYLE)).Get_Values (t_red, t_green, t_blue);  // fill color
    cgm_col_color(acolour,1,to255(t_red),to255(t_green),to255(t_blue));
    m_textcolr(acolour); //edge color as specified, or black for default
    }
  }
}

void Am_Save_To_GRAPHIC (Am_Object object, Am_String graphmode)
{
  if (!object.Valid ())
    Am_Error ("** Am_Save_To_GRAPHIC called with NULL object\n");
// RALCGM variables

Enum type;
RGBcolour rgbcol1;
RGBcolour rgbcol2;
Colour col1,col2,col3;
Code aspect[17];
Enum value[17];
Point a,b,c,d,e,f;  //these are floats, .x,.y
Point pline[10];   //for small lists of points
  int gr_w_var, gr_h_var, gr_t_var, gr_l_var; 
  int gr_dim;
  int itemp;
  int ang1,ang2,ang3;
  float ftemp;
  float fang1,fang2,fang3;
  float t_red, t_green,t_blue;  
  int cx,cy, sx,sy,ix,iy,ex,ey;
  float fx,fy;
  float cradius;
  char *cgm_env;
  static char dev_options[80]="";

// there might not be anything to draw!
  Am_Value_List components;
  components = object.Get (Am_GRAPHICAL_PARTS);
  if(components.Empty())  
    return;

// convert from the external graphmode string to RALCGM enumerated type
// default to POSTSCRIPT.  Pick up POSTSCRIPT options from environment, if 
// any after checking for all other types
  type = POSTSCRIPT;
  char *s = graphmode;
  if (strcmp(s,"CHARACTER")==0)type = CHARACTER;
  if (strcmp(s,"CLEAR_TEXT")==0) type = CLEAR_TEXT;
  if (strcmp(s,"BINARY")==0) type = BINARY;
  if (strcmp(s,"HPGL")==0){
     type = HPGL;
     cgm_env = (char *)getenv( "CGMHPGLOPT" );
     if( cgm_env != NULL )
         strcpy(dev_options,cgm_env);
     else
         dev_options[0]='\0';
     HPGLoptions( dev_options );
     free(cgm_env);
     }
  if(type == POSTSCRIPT){
     cgm_env = (char *)getenv( "CGMPSOPT" );
     if( cgm_env != NULL )
         strcpy(dev_options,cgm_env);
     else
         dev_options[0]='\0';
     PSoptions( dev_options );
     free(cgm_env);
     }

// pop up a dialog to get the name of the output file
  Am_String fn = Am_Get_Input_From_Dialog (
    Am_Value_List()
      .Add("Graphics Output Type is")
      .Add(graphmode)
      .Add("Output filename = "), 
    "CGM.OUT", 
    Am_AT_CENTER_SCREEN,
    Am_AT_CENTER_SCREEN,
    false);
  if(fn == Am_No_Value)return;  //user aborted with CANCEL
  char *filechar = fn;
//
// set up the CGM draw area, basically we just need to know how big the
// top object is and the background color
//

  cgminit(type, filechar);
  cgmverbose = FALSE;
// start of metafile description
  m_begmf("Amulet");
  m_mfversion ( 1 );
  m_mfdesc("Graphics output from Amulet");
  m_vdctype(REAL);
  cgm_rgb_color(rgbcol1,0,0,0);
  cgm_rgb_color(rgbcol2,255,255,255);
  m_colrvalueext(rgbcol1, rgbcol2);
  char *fontl[12] = { "Courier",   "Courier-Bold", "Courier-Oblique", "Courier-BoldOblique",
                      "Times-Roman", "Times-Bold", "Times-Italic", "Times-BoldItalic", 
                      "Helvetica", "Helvetica-Bold", "Helvetica-Oblique", "Helvetica-BoldOblique" };
  m_fontlist( 12, fontl );
  m_begmfdefaults();
    aspect[0] = ALL; value[0] = INDIV;
    m_asf( 1, aspect, value );
    switch ( type )
    {
      case CHARACTER:
         m_vdcrealprec( 16, -10, -16, ALLOWED );
         break;
      default:
         break;
    }
  m_endmfdefaults();
// end of of metafile description
// start of Picture.  Amulet graphics output can only be one picture, even
//    in a CGM file

  m_begpic("Amulet Output");
  m_colrmode(DIRECT);
// set picture size
  cgm_set_abspoint(a, 0, 0);
  cgm_set_abspoint(b,object.Get(Am_WIDTH),object.Get(Am_HEIGHT));
  cgm_save_y = object.Get(Am_HEIGHT); // Y runs in opposite direction in CGM
  m_vdcext(a,b);
//
  Am_Style (object.Get(Am_FILL_STYLE)).Get_Values (t_red, t_green, t_blue); // each 0<->1
//
// first color allocated is background color
//
   cgm_rgb_color(rgbcol1,to255(t_red),to255(t_green),to255(t_blue));
   m_backcolr(rgbcol1);
//
// that's it for the introductory stuff, now the actual picture itself
//
   m_begpicbody();
//
// now cycle through all OTHER components
//
    components.Start ();
    Am_Object current;
    for (components.Start(); !components.Last(); components.Next()) {
      current = components.Get ();
      if (current.Is_Instance_Of (Am_Graphical_Object)) {
//
//  rectangle code.  roundtangle rendered as a rectangle for
//  now, otherwise it would be a very complicated hybrid type
//  composed of one polygon (rectangle with corners hacked off)
//  and four half circles (the round part of the roundtangle).
//
        if (current.Is_Instance_Of (Am_Rectangle)  || 
            current.Is_Instance_Of (Am_Roundtangle)  ){
          set_gr_fill_edge_properties(current, 0); // self explanatory
          gr_w_var = current.Get(Am_WIDTH); // size and position of rectangle
          gr_h_var = current.Get(Am_HEIGHT);
          gr_l_var = current.Get(Am_LEFT);
          gr_t_var = current.Get(Am_TOP);
	  cgm_set_point(a,gr_l_var,gr_t_var);
	  cgm_set_point(b,gr_l_var + gr_w_var - 1, gr_t_var - gr_h_var + 1);
          m_rect(a ,b);
          }
//
//  CGM doesn't know what an Arrow line is, use Line instead
//
        if (current.Is_Instance_Of (Am_Line)     ||
            current.Is_Instance_Of (Am_Arrow_Line) ){
          set_gr_fill_edge_properties(current, 2); // self explanatory
	  cgm_set_point(pline[0],current.Get(Am_X1),current.Get(Am_Y1));
	  cgm_set_point(pline[1],current.Get(Am_X2),current.Get(Am_Y2));
          m_line(2, pline);
          }
//
// for polygons we need to extract the Am_Point_List, which means that
// first we need to find out how many there are, then malloc enough space
// for them, then extract them all with a start/next if != last
//
// This gets a bit messy.  An Amulet polygon is closed if the last and 
// first points are the same.  If they aren't, then you can still have a 
// fill on the line connecting them.  For CGM, if it isn't filled, and
// the end points aren't the same, it is a polyLINE.  Here is the full 
// mapping
//
// Amulet    Fill     CGM
// 1 == N     -       polygon  (drop Nth point)    CASE A
// 1 != N     y       polygon  (add N+1 th point)  CASE B
// 1 != N     n       polyline (keep Nth point)    CASE C
//

        if (current.Is_Instance_Of (Am_Polygon)){
          set_gr_fill_edge_properties(current, 0); // self explanatory
	  Am_Point_List thelist = current.Get(Am_POINT_LIST);
	  int howmany = thelist.Length();
	  Point *plist, *psavelist;
	  psavelist  = (Point *)malloc( 2 * (howmany+1) * sizeof(float) );
	  plist = psavelist;
	  if(plist == NULL)
               Am_Error ("** Am_Save_To_GRAPHIC ran out of memory\n");
	  for (thelist.Start(); !thelist.Last(); thelist.Next()){
	     thelist.Get(fx,fy);
	     plist->x = fx;
	     plist->y = ((float)cgm_save_y) - fy;
	     plist++;
             }
//
//  now, which CASE do we use
//
	  plist--;
	  if( (plist->x == psavelist->x) && (plist->y == psavelist->y)) {  //CASE A
	    howmany--;
	    m_polygon(howmany, psavelist);
            }
          else {
            if(current.Peek(Am_FILL_STYLE).Exists() &&
               current.Get(Am_FILL_STYLE) != Am_No_Style){  //CASE B
	       howmany++;
	       plist++;
	       plist->x = psavelist->x;
	       plist->y = psavelist->y;
	       m_polygon(howmany, psavelist);
               }
	    else { //CASE C
	       m_line(howmany, psavelist);
	       }
	    }
	  free(psavelist); // no memory leaks!!!!
          }
//
//  for arcs we need to get out Am_ANGLE1 and Am_ANGLE2 and convert
//  to start/center/end coordinates.  Since height can be different than
//  width always use elliptical CGM functions instead of circular ones.
//
//  if ANGLE2 = 360 or -360, use the ellipse function instead of the arc
//
        if (current.Is_Instance_Of (Am_Arc)){
          if(current.Get(Am_FILL_STYLE) == Am_No_Style)
               gr_dim = 1;
          else 
               gr_dim = 2;
          set_gr_fill_edge_properties(current, 0); // self explanatory
          gr_w_var = current.Get(Am_WIDTH); // size and position of rectangle
          gr_h_var = current.Get(Am_HEIGHT);
          gr_l_var = current.Get(Am_LEFT);
          gr_t_var = current.Get(Am_TOP);
// set center and two determining points on ellipse.  Since Amulet ellipses
// are always symmetric on X/Y axis, use one value on each axis

	  cgm_set_point(a, gr_l_var + (gr_w_var)/2 , gr_t_var - (gr_h_var)/2);
	  cgm_set_point(b, gr_l_var, gr_t_var - (gr_h_var)/2);
	  cgm_set_point(c, gr_l_var + (gr_w_var)/2 , gr_t_var);

          cradius = ((float) gr_w_var)/2.0;
//
//  Angles are different.  In Amulet it is start at ANG1 then move to ANG1+ANG2
//  In CGM it is draw from ANG1 to ANG2. Also, it only draws in one direction
//  so if Amulet's ANG2 is negative, need to swap the points

          ang1 = current.Get(Am_ANGLE1);
          ang3 = current.Get(Am_ANGLE2);
	  if(ang3 < 0){
	      ang2 = ang1;
	      ang1 = ang1 + ang3;
	      }
	  else
	      ang2 = ang1 + ang3;
          fang1 = ((float) ang1) * 2.0 * 3.134159 / 360.0;
          fang2 = ((float) ang2) * 2.0 * 3.134159 / 360.0;
          d.x = cos(fang1);  // generate start, intermediate
          d.y = sin(fang1);  // and end points on the arc.
          e.x = cos(fang2);  // These specify directions from the origin=center
          e.y = sin(fang2);  // a line out intersects arc and draws from there
	  if( (ang3 == 360) || (ang3 == -360) ){
             m_ellipse(a,b,c);
	     }
	  else {
            if(gr_dim == 1)
                m_elliparc(a,b,c,d,e);
            else
                m_elliparcclose(a,b,c,d,e,PIE); //Amulet uses PIE closure
            }
          }
	  
//
// for text we have to worry about which of the three fonts to use
// for size, work from the Am_HEIGHT slot rather than the Am_Font_Size_Flag
// charexpan factor is empirically determined :-( and set in gr_fill_edge
//
        if (current.Is_Instance_Of (Am_Text)){
          set_gr_fill_edge_properties(current, 1); // sets text color and font only
          gr_h_var = current.Get(Am_HEIGHT);
          gr_w_var = current.Get(Am_WIDTH);
	  ftemp = ((float) gr_h_var);
	  m_charheight(ftemp);
          gr_l_var = current.Get(Am_LEFT);
          gr_t_var = current.Get(Am_TOP);
	  cgm_set_point(a, gr_l_var , gr_t_var - gr_h_var + 1);
          Am_String s_temp = current.Get(Am_TEXT);
	  char *ss_temp = s_temp;
	  m_text(a,FINAL,ss_temp);
          }
//
//  Bitmap code.  Bitmaps don't have any pieces to worry about
//  other than their size and position.
//

        if (current.Is_Instance_Of (Am_Bitmap)){
          gr_w_var = current.Get(Am_WIDTH); // size and position of bitmap
          gr_h_var = current.Get(Am_HEIGHT);
          gr_l_var = current.Get(Am_LEFT);
          gr_t_var = current.Get(Am_TOP);
// three control points of image rectangle = UL, LR, UR
          cgm_set_point(a,gr_l_var,gr_t_var); 
          cgm_set_point(b,gr_l_var + gr_w_var - 1,gr_t_var - gr_h_var + 1); 
          cgm_set_point(c,gr_l_var + gr_w_var - 1,gr_t_var); 
          Am_Image_Array image = current.Get(Am_IMAGE); //get image from bitmap
          int width, height;
          image.Get_Size(width,height);
          unsigned char * storage = new unsigned char[width * height * 3];
          if (image.Get_RGB_Image(storage,true)) {
            m_cellarrayc ( a, b, c, width, height, storage );
            }
          else {
               cout << "Get_RGB_Image not implemented on this platform." << endl;
            }
          delete [] storage;
          }
      }
    }
// now close up the file
    m_endpic();
    m_endmf();
}
